import threading
import time

from snisp.agent import Agent


"""
This recipe will run until all of the Markets in your Command Ship's System
have a Probe DOCKED or IN_ORBIT at each Market.

The recipe can be run in it's own thread and it requires no monitoring,
which is ideal for running as a sub-script for a larger game.

Each new iteration of agent.fleet.probes() will return the most up-to-date
list of Probes. Meaning, if you purchase additional Probes in a seperate
thread in the time between iterations, the Probes will automatically
show up here.
"""


def main(*, agent_symbol, faction='', email='', token=''):
    # Get your agent
    agent = Agent(
        symbol=agent_symbol, faction=faction, email=email, token=token
    )

    # Fire off a thread
    # The thread will run until either all Markets or probed
    # or the program exits
    t = threading.Thread(target=run, args=(agent,), daemon=True)
    t.start()


def run(agent):
    # Get your command ship
    command_ship = next(iter(agent.fleet))

    # Get all of the Markets in your Command Ship's System
    markets = list(command_ship.markets)

    # Empty set for tracking probed Markets
    markets_with_probes = {}

    # Set of Probes that are at their Market
    probes_at_markets = {}

    while True:
        next_arrival = 60 * 15  # Fifteen minute baseline
        for probe in agent.fleet.probes():
            if probe.symbol in probes_at_markets:
                # Already at a Market
                continue
            if probe.nav.status == 'IN_TRANSIT':
                # Probe may already be in transit to a Market from
                # a previous iteration or function
                if probe.arrival < next_arrival:
                    # Set the timer to reset once this Probe arrives at
                    # it's destination
                    next_arrival = probe.arrival
                continue
            if probe.at_market:
                # The Probe is either at a Market that already has a probe,
                # which means it needs to move on to another Market; or,
                # it's at a Market that does not have a Probe listed at it,
                # which means it's safe to stay and should notify other Probes
                if probe.nav.waypoint_symbol not in markets_with_probes:
                    markets_with_probes.add(probe.nav.waypoint_symbol)
                    # Mark the Probe as at a Market
                    probes_at_markets.add(probe.symbol)
                    continue

            # At this point we know the Probe is not IN_TRANSIT and is not
            # at a Market that already has a Probe at it
            # Next thing to do is find the closest Market to the Probe
            next_market = probe.closest(markets)
            while next_market.symbol in markets_with_probes:
                # Drain the Markets pool of already known Probed Markets
                _ = markets.remove(next_market)
                if not markets:
                    # No remaining Markets! All Markets have been Probed
                    # or have a Probe heading their way
                    return
                next_market = probe.closest(markets)

            # Move the Probe on to the new Market location
            probe.update_flight_mode('CRUISE')
            # .navigate will *not* block if the Probe is not already IN_TRANSIT
            # The check for IN_TRANSIT at the top of the loop confirmed this
            probe.navigate(next_market)

            # Mark the Market as Probed
            markets_with_probes.add(next_market.symbol)

            # Mark the Probe as at a Market
            probes_at_markets.add(probe.symbol)

            # Remove the Market from the list
            _ = markets.remove(next_market)
            if not markets:
                # No remaining Markets! All Markets have been Probed
                # or have a Probe heading their way
                return

        # Wait either the default 15 minutes or when the next Probe is set to
        # arrive at its destination
        time.sleep(next_arrival)



import copy
import httpx
import json
import os
import pytest

import snisp

from . import DATA_DIR, GenericSideEffect


class TestShipyards:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_repr(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        assert repr(ship.shipyards) == f'Shipyards({ship.agent!r}, ' \
                                       f'{ship.location})'

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_iter(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')

        with shipyards_side_effect:
            for shipyard in ship.shipyards:
                assert any(i.symbol == 'SHIPYARD' for i in shipyard.traits)
            shipyards = list(ship.shipyards)
            assert len(shipyards) == 1

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_autopurchase(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        for ship in shipyards_data['data']['ships']:
            if ship['type'] == 'SHIP_PROBE':
                ship['purchasePrice'] = 100_000
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoint.json'), encoding='utf8')
        )
        waypoints_data['data']['traits'][0]['symbol'] = 'SHIPYARD'
        waypoints_data['data'] = [waypoints_data['data']]
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_data['data']['nav']['status'] = 'IN_TRANSIT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        agent_data = json.load(
            open(os.path.join(DATA_DIR, 'agent_data.json'), encoding='utf8')
        )
        agent_side_effect = GenericSideEffect(agent_data)
        agent_route = respx_mock.get('/my/agent')
        agent_route.side_effect = agent_side_effect

        ships_data = json.load(
            open(
                os.path.join(DATA_DIR, 'fleet_list_ships.json'),
                encoding='utf8'
            )
        )
        fleet_side_effect = FleetSideEffect(ships_data)
        fleet_route = respx_mock.get('/my/ships')
        fleet_route.side_effect = fleet_side_effect

        ship_purchase_data = json.load(
            open(
                os.path.join(DATA_DIR, 'ship_purchase.json'),
                encoding='utf8'
            )
        )
        respx_mock.post('/my/ships').mock(
            return_value=httpx.Response(200, json=ship_purchase_data)
        )

        with fleet_side_effect as tmp:
            tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')

            assert not ship.shipyards.autopurchase(ship_type='SHIP_PROBE')
            assert not ship.shipyards.autopurchase(
                ship_type='SHIP_REFINING_FREIGHTER'
            )

        with agent_side_effect as agent_tmp, fleet_side_effect as fleet_tmp:
            agent_tmp.data['data']['credits'] = 500_000
            agent = snisp.agent.Agent(symbol='testing', faction='testing')
            fleet_tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            fleet_tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            ship = agent.fleet('TEST_SHIP_SYMBOL')
            transactions = ship.shipyards.autopurchase(
                ship_type='SHIP_PROBE', max_units=2
            )
            assert len(transactions) == 2
            for transaction in transactions:
                assert transaction.frame.symbol == 'FRAME_PROBE'

        with fleet_side_effect as tmp:
            tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(snisp.exceptions.SpaceAttributeError):
                ship.shipyards.autopurchase(ship_type='INVALID')

        with agent_side_effect as agent_tmp, fleet_side_effect as fleet_tmp:
            agent_tmp.data['data']['credits'] = 500_000
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(snisp.exceptions.NoShipAtLocationError):
                list(ship.shipyards.autopurchase(ship_type='SHIP_PROBE'))


class TestShipyard:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_data(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-CLOSESTWAYPOINT'
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['symbol'] = 'TEST-SYSTEM-CLOSESTWAYPOINT'
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/'
            'TEST-SYSTEM-CLOSESTWAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        respx_mock.get('/my/ships/TEST_SHIP_SYMBOL').mock(
            return_value=httpx.Response(200, json=ship_data)
        )
        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        shipyard = next(iter(ship.shipyards))

        assert shipyard.data.symbol == 'TEST-SYSTEM-CLOSESTWAYPOINT'
        assert shipyard.data.to_dict() == shipyards_data['data']

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_purchase(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-CLOSESTWAYPOINT'
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['symbol'] = 'TEST-SYSTEM-CLOSESTWAYPOINT'
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        ships_data = json.load(
            open(
                os.path.join(DATA_DIR, 'fleet_list_ships.json'),
                encoding='utf8'
            )
        )
        for ship in ships_data['data']:
            ship['nav']['waypointSymbol'] = 'TEST-SYSTEM-INVALID'
            ship['nav']['status'] = 'IN_ORBIT'

        respx_mock.get(
                '/my/ships', params={'page': 1, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json=ships_data)
        )

        respx_mock.get(
                '/my/ships', params={'page': 2, 'limit': 20}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        shipyard = next(iter(ship.shipyards))

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            shipyard.purchase('INVALID')

        with ship_side_effect as tmp:
            tmp.data['data']['nav']['waypointSymbol'] = 'NOT-SYSTEM-WAYPOINT'
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            with pytest.raises(snisp.exceptions.NoShipAtLocationError):
                list(shipyard.purchase('SHIP_ORE_HOUND'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_transactions(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        ships_data = json.load(
            open(
                os.path.join(DATA_DIR, 'fleet_list_ships.json'),
                encoding='utf8'
            )
        )
        fleet_side_effect = FleetSideEffect(ships_data)
        fleet_route = respx_mock.get('/my/ships')
        fleet_route.side_effect = fleet_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        shipyard = next(iter(ship.shipyards))

        with fleet_side_effect as tmp:
            tmp.data['data'][0]['nav']['status'] = 'IN_ORBIT'
            tmp.data['data'][0]['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'  # noqa: E501
            transactions = list(shipyard.transactions())
            assert len(transactions) == 1
            assert transactions[0].price == 0

        with fleet_side_effect as tmp:
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            shipyard = snisp.shipyards.Shipyard(self.agent, shipyard.to_dict())
            with pytest.raises(snisp.exceptions.NoShipAtLocationError):
                list(shipyard.transactions())


class TestShipyardShip:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_available_ships(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        ship = self.agent.fleet('TEST_SHIP_SYMBOL')
        shipyard = next(iter(ship.shipyards))
        ships = list(shipyard.available_ships())
        assert len(ships) == 1

        ships = list(shipyard.available_ships('SHIP_EXPLORER'))
        assert len(ships) == 0

        with pytest.raises(snisp.exceptions.SpaceAttributeError):
            ships = list(shipyard.available_ships('invalid'))

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_purchase(self, respx_mock):
        shipyards_data = json.load(
            open(os.path.join(DATA_DIR, 'shipyards.json'), encoding='utf8')
        )
        shipyards_data['data']['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data = json.load(
            open(os.path.join(DATA_DIR, 'waypoints.json'), encoding='utf8')
        )
        waypoints_data['data'][0]['symbol'] = 'TEST-SYSTEM-WAYPOINT'
        waypoints_data['data'][0]['traits'][0]['symbol'] = 'SHIPYARD'
        shipyards_side_effect = ShipyardsSideEffect(
            data=shipyards_data, waypoints=waypoints_data
        )
        shipyards_route = respx_mock.get(
            '/systems/TEST-SYSTEM/waypoints/TEST-SYSTEM-WAYPOINT/shipyard'
        )
        shipyards_route.side_effect = shipyards_side_effect
        waypoints_route = respx_mock.get('/systems/TEST-SYSTEM/waypoints')
        waypoints_route.side_effect = shipyards_side_effect.waypoints_side_effect  # noqa: E501

        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        ship_data['data']['nav']['systemSymbol'] = 'TEST-SYSTEM'
        ship_data['data']['nav']['waypointSymbol'] = 'TEST-SYSTEM-WAYPOINT'
        ship_side_effect = GenericSideEffect(ship_data)
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = ship_side_effect

        agent_data = json.load(
            open(os.path.join(DATA_DIR, 'agent_data.json'), encoding='utf8')
        )
        agent_side_effect = GenericSideEffect(agent_data)
        agent_route = respx_mock.get('/my/agent')
        agent_route.side_effect = agent_side_effect

        ship_purchase_data = json.load(
            open(
                os.path.join(DATA_DIR, 'ship_purchase.json'),
                encoding='utf8'
            )
        )
        respx_mock.post('/my/ships').mock(
            return_value=httpx.Response(200, json=ship_purchase_data)
        )

        with agent_side_effect as tmp:
            tmp.data['data']['credits'] = 500_000
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            shipyard = next(iter(ship.shipyards))
            shipyard_ship = next(iter((shipyard.available_ships())))
            purchase = shipyard_ship.purchase()
            assert purchase.frame.symbol == 'FRAME_PROBE'

        with agent_side_effect as tmp:
            tmp.data['data']['credits'] = 500
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            shipyard = next(iter(ship.shipyards))
            shipyard_ship = next(iter((shipyard.available_ships())))
            assert not shipyard_ship.purchase(buffer=500_000)


class ShipyardsSideEffect:

    def __init__(self, *, data, waypoints):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.orig_waypoints = waypoints
        self.waypoints = copy.deepcopy(self.orig_waypoints)

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': {}})
        return httpx.Response(200, json=self.data)

    def waypoints_side_effect(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        traits = request.url.params.get('traits')
        output = []
        for wp in self.waypoints['data']:
            if any(i['symbol'] == traits for i in wp['traits']):
                output.append(wp)
        self.waypoints['data'] = output
        return httpx.Response(200, json=self.waypoints)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.waypoints = copy.deepcopy(self.orig_waypoints)


class FleetSideEffect:

    def __init__(self, data):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': {}})
        return httpx.Response(200, json=self.data)

    def __enter__(self):
        self.reset()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)



from datetime import datetime, timezone


class SpaceBaseException(Exception):
    pass


class ClientError(SpaceBaseException):

    def __init__(self, msg=None, data=None):
        msg = msg if msg is not None else 'Generic Client Exception'
        super().__init__(msg)
        self.data = data
        self.__raised_at = datetime.now(timezone.utc)

    @property
    def raised_at(self):
        return self.__raised_at


class SpaceAttributeError(SpaceBaseException):
    pass


class SpaceUserError(SpaceBaseException):
    pass


class NoShipAtLocationError(SpaceBaseException):
    pass


# General Error Codes
class CooldownConflictError(SpaceBaseException):
    error_code = 4000


class WaypointNoAccessError(SpaceBaseException):
    error_code = 4001


# Account Error Codes
class AccountException(SpaceBaseException):
    pass


class TokenEmptyError(AccountException):
    error_code = 4100


class TokenMissingSubjectError(AccountException):
    error_code = 4101


class TokenInvalidSubjectError(AccountException):
    error_code = 4102


class MissingTokenRequestError(AccountException):
    error_code = 4103


class InvalidTokenRequestError(AccountException):
    error_code = 4104


class InvalidTokenSubjectError(AccountException):
    error_code = 4105


class AccountNotExistsError(AccountException):
    error_code = 4106


class AgentNotExistsError(AccountException):
    error_code = 4107


class AccountHasNoAgentError(AccountException):
    error_code = 4108


class RegisterAgentExistsError(AccountException):
    error_code = 4109


class RegisterAgentSymbolReservedError(AccountException):
    error_code = 4110


class RegisterAgentConflictSymbolError(AccountException):
    error_code = 4111


# Ship Error Codes
class ShipException(SpaceBaseException):
    pass


class NavigateInTransitError(ShipException):
    error_code = 4200


class NavigateInvalidDestinationError(ShipException):
    error_code = 4201


class NavigateOutsideSystemError(ShipException):
    error_code = 4202


class NavigateInsufficientFuelError(ShipException):
    error_code = 4203


class NavigateSameDestinationError(ShipException):
    error_code = 4204


class ShipExtractInvalidWaypointError(ShipException):
    error_code = 4205


class ShipExtractPermissionError(ShipException):
    error_code = 4206


class ShipJumpNoSystemError(ShipException):
    error_code = 4207


class ShipJumpSameSystemError(ShipException):
    error_code = 4208


class ShipJumpMissingModuleError(ShipException):
    error_code = 4210


class ShipJumpNoValidWaypointError(ShipException):
    error_code = 4211


class ShipJumpMissingAntimatterError(ShipException):
    error_code = 4212


class ShipInTransitError(ShipException):
    error_code = 4214


class ShipMissingSensorArraysError(ShipException):
    error_code = 4215


class PurchaseShipCreditsError(ShipException):
    error_code = 4216


class ShipCargoExceedsLimitError(ShipException):
    error_code = 4217


class ShipCargoMissingError(ShipException):
    error_code = 4218


class ShipCargoUnitCountError(ShipException):
    error_code = 4219


class ShipSurveyVerificationError(ShipException):
    error_code = 4220


class ShipSurveyExpirationError(ShipException):
    error_code = 4221


class ShipSurveyWaypointTypeError(ShipException):
    error_code = 4222


class ShipSurveyOrbitError(ShipException):
    error_code = 4223


class ShipSurveyExhaustedError(ShipException):
    error_code = 4224


class ShipRefuelDockedError(ShipException):
    error_code = 4225


class ShipRefuelInvalidWaypointError(ShipException):
    error_code = 4226


class ShipCargoFullError(ShipException):
    error_code = 4228


class ShipJumpFromGateToGateError(ShipException):
    error_code = 4229


class WaypointChartedError(ShipException):
    error_code = 4230


class ShipTransferShipNotFound(ShipException):
    error_code = 4231


class ShipTransferAgentConflict(ShipException):
    error_code = 4232


class ShipTransferSameShipConflict(ShipException):
    error_code = 4233


class ShipTransferLocationConflict(ShipException):
    error_code = 4234


class WarpInsideSystemError(ShipException):
    error_code = 4235


class ShipNotInOrbitError(ShipException):
    error_code = 4236


class ShipInvalidRefineryGoodError(ShipException):
    error_code = 4237


class ShipInvalidRefineryTypeError(ShipException):
    error_code = 4238


class ShipMissingRefineryError(ShipException):
    error_code = 4239


class ShipMissingSurveyorError(ShipException):
    error_code = 4240


class ShipMissingWarpDriveError(ShipException):
    error_code = 4241


class ShipMissingMineralProcessorError(ShipException):
    error_code = 4242


class ShipMissingMiningLasersError(ShipException):
    error_code = 4243


class ShipNotDockedError(ShipException):
    error_code = 4244


class PurchaseShipNotPresentError(ShipException):
    error_code = 4245


class ShipMountNoShipyardError(ShipException):
    error_code = 4246


class ShipMissingMountError(ShipException):
    error_code = 4247


class ShipMountInsufficientCreditsError(ShipException):
    error_code = 4248


class ShipMissingPowerError(ShipException):
    error_code = 4249


class ShipMissingSlotsError(ShipException):
    error_code = 4250


class ShipMissingMountsError(ShipException):
    error_code = [4251, 4227]


class ShipMissingCrewError(ShipException):
    error_code = 4252


class ShipExtractDestabilizedError(ShipException):
    error_code = 4253


class ShipJumpInvalidOriginError(ShipException):
    error_code = 4254


class ShipJumpInvalidWaypointError(ShipException):
    error_code = 4255


class ShipJumpOriginUnderConstructionError(ShipException):
    error_code = 4256


class ShipMissingGasProcessorError(ShipException):
    error_code = 4257


class ShipMissingGasSiphonsError(ShipException):
    error_code = 4258


class ShipSiphonInvalidWaypointError(ShipException):
    error_code = 4259


class ShipSiphonPermissionError(ShipException):
    error_code = 4260


class WaypointNoYieldError(ShipException):
    error_code = 4261


class ShipJumpDestinationUnderConstructionError(ShipException):
    error_code = 4262


# Contract Error Codes
class ContractException(SpaceBaseException):
    pass


class AcceptContractNotAuthorizedError(ContractException):
    error_code = 4500


class AcceptContractConflictError(ContractException):
    error_code = 4501


class FulfillContractDeliveryError(ContractException):
    error_code = 4502


class ContractDeadlineError(ContractException):
    error_code = 4503


class ContractFulfilledError(ContractException):
    error_code = 4504


class ContractNotAcceptedError(ContractException):
    error_code = 4505


class ContractNotAuthorizedError(ContractException):
    error_code = 4506


class ShipDeliverTermsError(ContractException):
    error_code = 4508


class ShipDeliverFulfilledError(ContractException):
    error_code = 4509


class ShipDeliverInvalidLocationError(ContractException):
    error_code = 4510


class ExistingContractError(ContractException):
    error_code = 4511


# Market Error Codes
class MarketException(SpaceBaseException):
    pass


class MarketInsufficientCredits(MarketException):
    error_code = 4600


class MarketTradeNoPurchaseError(MarketException):
    error_code = 4601


class MarketTradeNotSoldError(MarketException):
    error_code = 4602


class MarketNotFoundError(MarketException):
    error_code = 4603


class MarketTradeUnitLimitError(MarketException):
    error_code = 4604


# Faction Error Codes
class FactionException(SpaceBaseException):
    pass


class WaypointNoFactionError(FactionException):
    error_code = 4700


# Construction Error Code
class ConstructionException(SpaceBaseException):
    pass


class ConstructionMaterialNotRequired(ConstructionException):
    error_code = 4800


class ConstructionMaterialFulfilled(ConstructionException):
    error_code = 4801


class ShipConstructionInvalidLocationError(ConstructionException):
    error_code = 4802


error_codes = {
    # TODO: hiearchy
    # General Error Codes
    4000: CooldownConflictError,
    4001: WaypointNoAccessError,

    # Account Error Codes
    4100: TokenEmptyError,
    4101: TokenMissingSubjectError,
    4102: TokenInvalidSubjectError,
    4103: MissingTokenRequestError,
    4104: InvalidTokenRequestError,
    4105: InvalidTokenSubjectError,
    4106: AccountNotExistsError,
    4107: AgentNotExistsError,
    4108: AccountHasNoAgentError,
    4109: RegisterAgentExistsError,
    4110: RegisterAgentSymbolReservedError,
    4111: RegisterAgentConflictSymbolError,

    # Ship Error Codes
    4200: NavigateInTransitError,
    4201: NavigateInvalidDestinationError,
    4202: NavigateOutsideSystemError,
    4203: NavigateInsufficientFuelError,
    4204: NavigateSameDestinationError,
    4205: ShipExtractInvalidWaypointError,
    4206: ShipExtractPermissionError,
    4207: ShipJumpNoSystemError,
    4208: ShipJumpSameSystemError,
    4210: ShipJumpMissingModuleError,
    4211: ShipJumpNoValidWaypointError,
    4212: ShipJumpMissingAntimatterError,
    4214: ShipInTransitError,
    4215: ShipMissingSensorArraysError,
    4216: PurchaseShipCreditsError,
    4217: ShipCargoExceedsLimitError,
    4218: ShipCargoMissingError,
    4219: ShipCargoUnitCountError,
    4220: ShipSurveyVerificationError,
    4221: ShipSurveyExpirationError,
    4222: ShipSurveyWaypointTypeError,
    4223: ShipSurveyOrbitError,
    4224: ShipSurveyExhaustedError,
    4225: ShipRefuelDockedError,
    4226: ShipRefuelInvalidWaypointError,
    4227: ShipMissingMountsError,  # Just a guess. It's a duplicate
    4228: ShipCargoFullError,
    4229: ShipJumpFromGateToGateError,
    4230: WaypointChartedError,
    4231: ShipTransferShipNotFound,
    4232: ShipTransferAgentConflict,
    4233: ShipTransferSameShipConflict,
    4234: ShipTransferLocationConflict,
    4235: WarpInsideSystemError,
    4236: ShipNotInOrbitError,
    4237: ShipInvalidRefineryGoodError,
    4238: ShipInvalidRefineryTypeError,
    4239: ShipMissingRefineryError,
    4240: ShipMissingSurveyorError,
    4241: ShipMissingWarpDriveError,
    4242: ShipMissingMineralProcessorError,
    4243: ShipMissingMiningLasersError,
    4244: ShipNotDockedError,
    4245: PurchaseShipNotPresentError,
    4246: ShipMountNoShipyardError,
    4247: ShipMissingMountError,
    4248: ShipMountInsufficientCreditsError,
    4249: ShipMissingPowerError,
    4250: ShipMissingSlotsError,
    4251: ShipMissingMountsError,
    4252: ShipMissingCrewError,
    4253: ShipExtractDestabilizedError,
    4254: ShipJumpInvalidOriginError,
    4255: ShipJumpInvalidWaypointError,
    4256: ShipJumpOriginUnderConstructionError,
    4257: ShipMissingGasProcessorError,
    4258: ShipMissingGasSiphonsError,
    4259: ShipSiphonInvalidWaypointError,
    4260: ShipSiphonPermissionError,
    4261: WaypointNoYieldError,
    4262: ShipJumpDestinationUnderConstructionError,

    # Contract Error Codes
    4500: AcceptContractNotAuthorizedError,
    4501: AcceptContractConflictError,
    4502: FulfillContractDeliveryError,
    4503: ContractDeadlineError,
    4504: ContractFulfilledError,
    4505: ContractNotAcceptedError,
    4506: ContractNotAuthorizedError,
    4508: ShipDeliverTermsError,
    4509: ShipDeliverFulfilledError,
    4510: ShipDeliverInvalidLocationError,
    4511: ExistingContractError,

    # Market Error Codes
    4600: MarketInsufficientCredits,
    4601: MarketTradeNoPurchaseError,
    4602: MarketTradeNotSoldError,
    4603: MarketNotFoundError,
    4604: MarketTradeUnitLimitError,

    # Faction Error Codes
    4700: WaypointNoFactionError,

    # Construction Error Code
    4800: ConstructionMaterialNotRequired,
    4801: ConstructionMaterialFulfilled,
    4802: ShipConstructionInvalidLocationError,
}



import httpx
import logging
import urllib.parse as urlparse

import snisp


logger = logging.getLogger(__name__)


# NOTE: Only waypoints work for now in DB

FUEL_STATIONS = {}


def get_fuel_stations(location):
    # Assumes the calling thread is already under a lock
    # Ideally, it will be the agent's lock
    return FUEL_STATIONS.get(location.system)


def insert_fuel_stations(location, fuel_stations):
    # Assumes the calling thread is already under a lock
    # Ideally, it will be the agent's lock
    FUEL_STATIONS[location.system] = list(fuel_stations)


def lookup(*args, **kwargs):
    # args: (client, url)
    # kwargs: {params}
    url = urlparse.urlsplit(args[1])
    parts = url.path.split('/')
    if parts[-1] == 'market':
        # Markets always need live data
        # Fucks up trade_volume otherwise
        return
    elif parts[-1] == 'waypoints':
        return waypoints_lookup(url, params=kwargs.get('params'))


def insert(response, *args, **kwargs):
    url = urlparse.urlsplit(args[1])
    parts = url.path.split('/')
    if parts[-1] == 'market':
        # Markets always need live data
        # Fucks up trade_volume otherwise
        return
    elif parts[-1] == 'waypoints':
        return waypoints_insert(response, url, params=kwargs.get('params'))


def waypoints_lookup(url, *, params=None):
    parts = url.path.split('/')
    params = params if params is not None else {}
    system = parts[-2]
    page = int(params.get('page', 0))
    page_limit = int(params.get('limit', 20))
    traits = params.get('traits')
    types = params.get('type')
    if data := snisp.database.get_waypoints(
        system=system,
        page_limit=page_limit,
        page=page,
        traits=traits,
        types=types,
    ):
        request = httpx.Request('GET', url.path, params=params)
        return httpx.Response(200, json=data, request=request)


def waypoints_insert(response, url, params=None):
    parts = url.path.split('/')
    system = parts[-2]
    params = params if params is not None else {}
    page = int(params.get('page', 0))
    page_limit = int(params.get('limit', 20))
    traits = params.get('traits')
    types = params.get('type')
    snisp.database.insert_waypoints(
        response.json(),
        system=system,
        page_limit=page_limit,
        page=page,
        traits=traits,
        types=types,
    )



import dateutil
import httpx
import json
import logging
import os

from collections import namedtuple
from datetime import datetime, timezone
from httpx import HTTPError

from snisp.exceptions import ClientError, SpaceUserError
from snisp.decorators import CachedRateLimiter


logger = logging.getLogger(__name__)
cached_rate_limiter = CachedRateLimiter()


UserData = namedtuple('UserData', ('symbol', 'faction', 'email', 'token'))


def cleanup_client(client):
    try:
        client.close()
    except Exception:
        pass


class SpaceClient(httpx.Client):

    def __init__(self, headers=None, token=None):
        if headers is None:
            headers = {'Content-Type': 'application/json'}
        self.testing = False
        if token is not None:
            headers['Authorization'] = f'Bearer {token}'
            if token == 'TESTING_TOKEN':
                self.testing = True
        super().__init__(
            headers=headers, base_url='https://api.spacetraders.io/v2'
        )

    def __del__(self):
        cleanup_client(self)

    def cleanup(self):
        cleanup_client(self)

    @cached_rate_limiter
    def delete(self, *args, **kwargs):
        response = super().delete(*args, **kwargs)
        raise_for_status(response)
        return response

    @cached_rate_limiter
    def get(self, *args, **kwargs):
        response = super().get(*args, **kwargs)
        raise_for_status(response)
        return response

    @cached_rate_limiter
    def options(self, *args, **kwargs):
        response = super().options(*args, **kwargs)
        raise_for_status(response)
        return response

    @cached_rate_limiter
    def patch(self, *args, **kwargs):
        response = super().patch(*args, **kwargs)
        raise_for_status(response)
        return response

    @cached_rate_limiter
    def post(self, *args, **kwargs):
        response = super().post(*args, **kwargs)
        raise_for_status(response)
        return response

    @cached_rate_limiter
    def put(self, *args, **kwargs):
        response = super().put(*args, **kwargs)
        raise_for_status(response)
        return response


def raise_for_status(response):  # pragma: no cover
    try:
        if isinstance(response, dict):
            # Assumes it's a cached response. Not great
            return
        response.raise_for_status()
    except HTTPError as e:
        logger.debug(f'FAILED RESPONSE: {response!r}')
        try:
            data = response.json()
        except json.decoder.JSONDecodeError:
            raise e
        if err := data.get('error'):
            msg = ' | '.join(
                f'{k.capitalize()}: {v}' for k, v in err.items()
            )
            err_data = err.get('data', err)
            if not err_data.get('code'):
                if code := err.get('code'):
                    err_data['code'] = code
            raise ClientError(msg, err_data)
        raise ClientError(str(e), data.get('data', data))


def load_user(
    *, symbol='', faction='', email='', token=''
):  # pragma: no cover
    symbol = symbol.upper().strip() if symbol else ''
    faction = faction.upper().strip() if faction else ''
    token = token.strip() if token else ''
    if symbol == 'TESTING' and faction == 'TESTING':
        return UserData(
            token='TESTING_TOKEN',
            symbol='TESTING',
            faction='TESTING',
            email='TESTING',
        )
    config_file = os.path.abspath(
        os.path.join(os.path.dirname(__file__), 'data', 'user_config.json')
    )
    last_login = str(datetime.now(timezone.utc))
    if os.path.isfile(config_file):
        user_data = json.load(open(config_file, 'r', encoding='utf8'))
        if not symbol and not token:
            if user := max(
                user_data.get('users', []),
                key=lambda x: dateutil.parser.parse(x['last_login'])
            ):
                user['last_login'] = last_login
                with open(config_file, 'w', encoding='utf8') as f:
                    json.dump(user_data, f, sort_keys=True, indent=2)
                return UserData(
                    token=user['token'],
                    symbol=user['symbol'],
                    faction=user['faction'],
                    email=user['email'],
                )
        for user in user_data.get('users', []):
            if user['symbol'].upper() == symbol or user['token'] == token:
                if faction and user['faction'] != faction:
                    raise SpaceUserError(
                        f'{user} already exists with a Faction of '
                        f'{user["faction"]!r}. {faction!r} is not compatible.'
                    )
                if email and user['email'] != email:
                    raise SpaceUserError(
                        f'{user} already exists but is associated with an'
                        f'email address of {user["email"]!r}.'
                    )
                user['last_login'] = last_login
                with open(config_file, 'w', encoding='utf8') as f:
                    json.dump(user_data, f, sort_keys=True, indent=2)
                return UserData(
                    token=user['token'],
                    symbol=user['symbol'],
                    faction=user['faction'],
                    email=user['email'],
                )
    else:
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        user_data = dict()
    if token:
        user_data.setdefault('users', []).append(
            {
                'symbol': symbol,
                'faction': faction,
                'token': token,
                'email': email,
                'last_login': last_login,
            }
        )
        with open(config_file, 'w', encoding='utf8') as f:
            json.dump(user_data, f, sort_keys=True, indent=2)
        return UserData(
            token=token,
            symbol=symbol,
            faction=faction,
            email=email,
        )
    if not symbol or not faction:
        raise SpaceUserError(
            '"symbol" and "faction" are required to create an account.'
        )
    try:
        session = SpaceClient(headers={})
        payload = {'symbol': symbol, 'faction': faction}
        if email is not None:
            payload['email'] = email
        response = session.post('/register', data=payload)
        token = response.json()['data']['token']
    except Exception as e:
        raise e
    finally:
        session.close()
    user_data.setdefault('users', []).append(
        {
            'symbol': symbol,
            'faction': faction,
            'token': token,
            'email': email,
            'last_login': last_login,
        }
    )
    with open(config_file, 'w', encoding='utf8') as f:
        json.dump(user_data, f, sort_keys=True, indent=2)
    return UserData(
        token=token,
        symbol=symbol,
        faction=faction,
        email=email,
    )



import atexit
import collections
import logging
import os
import threading

from snisp import database, utils
from snisp.client import SpaceClient, load_user
from snisp.contracts import Contracts
from snisp.factions import Factions
from snisp.fleet import Fleet
from snisp.systems import Systems


logger = logging.getLogger(__name__)


class Agent:

    """Agent represents your Player in SpaceTraders"""

    def __init__(self, *, symbol='', faction='', email='', token=''):
        database.setup()
        self.lock = threading.RLock()
        self.__systems = None
        user_data = load_user(
            symbol=symbol, faction=faction, email=email, token=token
        )
        self.__email = user_data.email
        self.__symbol = user_data.symbol
        self.__faction = user_data.faction
        self.__token = user_data.token
        self.__client = SpaceClient(token=self.token)
        atexit.register(self.client.cleanup)
        self.contracts = Contracts(self)
        self.fleet = Fleet(self)
        self.factions = Factions(self)
        self.dead_ships = dict()
        self.recent_transactions = collections.deque(maxlen=100)

    def __repr__(self):  # pragma: no cover
        cls = self.__class__.__name__
        output = f'{cls}(symbol={self.symbol!r}, faction={self.faction!r}'
        if self.email:
            return output + f', email={self.email!r})'
        else:
            return output + ')'

    def __del__(self):  # pragma: no cover
        try:
            self.client.close()
        except Exception:
            pass

    @property
    def client(self):
        """Property that returns the modified HTTPX Client"""
        return self.__client

    @property
    def data(self):
        """Your Agent's current data

        Returns:
            PlayerData
        """
        response = self.client.get('/my/agent')
        return PlayerData(self, response.json()['data'])

    @property
    def email(self):
        return self.__email

    @property
    def faction(self):
        return self.__faction

    @property
    def symbol(self):
        return self.__symbol

    @property
    def systems(self):
        if self.__systems is None:
            self.__systems = Systems(self)
        return self.__systems

    @property
    def token(self):
        return self.__token


class PlayerData(utils.AbstractJSONItem):

    """Your Agent's current data"""

    def __init__(self, agent, ship_data):
        self.agent = agent
        self._data = ship_data


def reset():  # pragma: no cover
    """Removes the user_config.json file if it exists"""
    config_file = os.path.abspath(
        os.path.join(os.path.dirname(__file__), 'data', 'user_config.json')
    )
    if os.path.isfile(config_file):
        try:
            os.remove(config_file)
        except Exception as e:
            raise Exception(
                f'Attempted to reset {config_file} but received {e!r}'
            )



import copy
import httpx
import json
import os
import pytest

from datetime import datetime, timedelta, timezone

import snisp

from . import DATA_DIR


class TestContracts:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_iter(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        respx_mock.get(
            '/my/contracts', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=contracts_data)
        )
        respx_mock.get(
            '/my/contracts', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        contracts = self.agent.contracts
        contracts = list(contracts)
        assert len(contracts) == 2
        assert contracts[0].fulfilled is True
        assert contracts[1].fulfilled is False

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_call(self, respx_mock):
        contract_data = json.load(
            open(os.path.join(DATA_DIR, 'contract.json'), encoding='utf8')
        )
        contract_data['data']['contract']['id'] = 'CONTRACT-ID'
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contracts_data['data'][-1]['id'] = 'CONTRACT-ID'
        respx_mock.get(
            '/my/contracts', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=contracts_data)
        )
        respx_mock.get(
            '/my/contracts', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )
        respx_mock.get(
            '/my/contracts/CONTRACT-ID'
        ).mock(
            return_value=httpx.Response(
                200, json={'data': contract_data['data']['contract']}
            )
        )

        contract = list(self.agent.contracts)[-1]
        assert not contract.accepted
        assert self.agent.contracts(contract.id) == contract

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_current(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        respx_mock.get(
            '/my/contracts', params={'page': 1}
        ).mock(
            return_value=httpx.Response(200, json=contracts_data)
        )
        respx_mock.get(
            '/my/contracts', params={'page': 2}
        ).mock(
            return_value=httpx.Response(200, json={'data': []})
        )

        contract = list(self.agent.contracts)[-1]
        assert contract == self.agent.contracts.current


class TestContract:

    agent = snisp.agent.Agent(symbol='testing', faction='testing')

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_accept(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        accept_contract_route = respx_mock.post(
            f'/my/contracts/{contracts_data["data"][-1]["id"]}/accept'
        )
        accept_contract_route.side_effect = contract_side_effect.accept

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['accepted'] = False
            assert not self.agent.contracts.current.accepted

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['accepted'] = True
            assert self.agent.contracts.current.accepted

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['accepted'] = False
            contract = self.agent.contracts.current
            contract.accept()
            assert contract.accepted

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_fulfilled(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        accept_contract_route = respx_mock.post(
            f'/my/contracts/{contracts_data["data"][-1]["id"]}/fulfill'
        )
        accept_contract_route.side_effect = contract_side_effect.fulfill

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['fulfilled'] = False
            assert not self.agent.contracts.current.fulfilled

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['fulfilled'] = True
            assert self.agent.contracts.current.fulfilled

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['fulfilled'] = False
            contract = self.agent.contracts.current
            contract.fulfill()
            assert contract.fulfilled

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_expired(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        assert self.agent.contracts.current.expired

        with contract_side_effect as tmp:
            expiration = datetime.now(timezone.utc) + timedelta(hours=1)
            tmp.data['data'][-1]['terms']['deadline'] = str(expiration)
            assert not self.agent.contracts.current.expired

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_refresh(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        respx_mock.get(
            f'/my/contracts/{contracts_data["data"][-1]["id"]}'
        ).mock(
            return_value=httpx.Response(
                200, json={'data': contracts_data['data'][-1]}
            )
        )

        contract = self.agent.contracts.current
        assert contract == contract.refresh()

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'INVALID'  # noqa: E501
            contract = self.agent.contracts.current
            assert contract != contract.refresh()

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_extractable(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            assert contract.extractable
            assert contract.__dict__['_data']['type'] == contract.type

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            with pytest.raises(AttributeError):
                contract.update_data_item('INVALID', '')
            with pytest.raises(AttributeError):
                contract.extractable = False

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'INVALID'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            assert not contract.extractable

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_siphonable(self, respx_mock):
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(contracts_data)
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'LIQUID_HYDROGEN'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            assert contract.siphonable

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'INVALID'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 0
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100  # noqa: E501
            contract = self.agent.contracts.current
            assert not contract.siphonable

    @pytest.mark.respx(base_url='https://api.spacetraders.io/v2')
    def test_deliver(self, respx_mock):
        ship_data = json.load(
            open(os.path.join(DATA_DIR, 'ship_info.json'), encoding='utf8')
        )
        contracts_data = json.load(
            open(os.path.join(DATA_DIR, 'contracts.json'), encoding='utf8')
        )
        contract_side_effect = ContractsSideEffect(
            contracts_data, ship_data=ship_data
        )
        contract_route = respx_mock.get('/my/contracts')
        contract_route.side_effect = contract_side_effect
        ship_route = respx_mock.get('/my/ships/TEST_SHIP_SYMBOL')
        ship_route.side_effect = contract_side_effect.ship
        deliver_contract_route = respx_mock.post(
            f'/my/contracts/{contracts_data["data"][-1]["id"]}/deliver'
        )
        deliver_contract_route.side_effect = contract_side_effect.deliver

        dock_data = json.load(
            open(os.path.join(DATA_DIR, 'dock.json'), encoding='utf8')
        )

        respx_mock.post('/my/ships/TEST_SHIP_SYMBOL/dock').mock(
            return_value=httpx.Response(200, json=dock_data)
        )

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0]['symbol'] = 'IRON_ORE'  # noqa: E501
            tmp.ship_data['data']['cargo']['inventory'][0]['units'] = 40
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            contract = self.agent.contracts.current
            contract.deliver(ship, 'IRON_ORE')
            assert contract.terms.deliver[0].units_required == 100
            assert contract.terms.deliver[0].units_fulfilled == 40
            assert ship.cargo.units == 0
            assert ship.cargo.inventory == []

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'INVALID'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0]['symbol'] = 'IRON_ORE'  # noqa: E501
            tmp.ship_data['data']['cargo']['inventory'][0]['units'] = 40
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            contract = self.agent.contracts.current
            assert contract.deliver(ship, 'IRON_ORE') is None

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0]['symbol'] = 'INVALID'  # noqa: E501
            tmp.ship_data['data']['cargo']['inventory'][0]['units'] = 40
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            contract = self.agent.contracts.current
            contract.deliver(ship, 'IRON_ORE')
            assert contract.terms.deliver[0].units_required == 100
            assert contract.terms.deliver[0].units_fulfilled == 0
            assert ship.cargo.units == 40

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 100
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0]['symbol'] = 'IRON_ORE'  # noqa: E501
            tmp.ship_data['data']['cargo']['inventory'][0]['units'] = 40
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            contract = self.agent.contracts.current
            contract.deliver(ship, 'IRON_ORE')
            assert contract.terms.deliver[0].units_required == 100
            assert contract.terms.deliver[0].units_fulfilled == 100
            assert ship.cargo.units == 40

        with contract_side_effect as tmp:
            tmp.data['data'][-1]['terms']['deliver'][0]['tradeSymbol'] = 'IRON_ORE'  # noqa: E501
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsRequired'] = 100
            tmp.data['data'][-1]['terms']['deliver'][0]['unitsFulfilled'] = 99
            tmp.ship_data['data']['cargo']['units'] = 40
            tmp.ship_data['data']['cargo']['inventory'][0]['symbol'] = 'IRON_ORE'  # noqa: E501
            tmp.ship_data['data']['cargo']['inventory'][0]['units'] = 40
            ship = self.agent.fleet('TEST_SHIP_SYMBOL')
            contract = self.agent.contracts.current
            contract.deliver(ship, 'IRON_ORE')
            assert contract.terms.deliver[0].units_required == 100
            assert contract.terms.deliver[0].units_fulfilled == 100
            assert ship.cargo.units == 39


class ContractsSideEffect:

    def __init__(self, data, *, ship_data=None):
        self.orig_data = data
        self.data = copy.deepcopy(self.orig_data)
        self.orig_ship_data = ship_data
        self.ship_data = copy.deepcopy(self.orig_ship_data)

    def __enter__(self):
        self.reset()
        return self

    def __call__(self, request, route):
        if int(request.url.params.get('page', 1)) > 1:
            return httpx.Response(200, json={'data': []})
        return httpx.Response(200, json=self.data)

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()

    def accept(self, request, route):
        return httpx.Response(200)

    def fulfill(self, request, route):
        self.data['data'][-1]['fulfilled'] = True
        return httpx.Response(
            200, json={'data': {'contract': self.data['data'][-1]}}
        )

    def ship(self, request, route):
        return httpx.Response(200, json=self.ship_data)

    def deliver(self, request, route):
        payload = json.loads(request.content.decode('utf8'))
        good = payload['tradeSymbol']
        units = int(payload['units'])
        output = []
        for item in self.ship_data['data']['cargo']['inventory']:
            if item['symbol'] == good:
                item['units'] -= units
                if item['units'] > 0:
                    output.append(item)
        self.ship_data['data']['cargo']['units'] -= units
        self.ship_data['data']['cargo']['inventory'] = output
        for item in self.data['data'][-1]['terms']['deliver']:
            if item['tradeSymbol'] == good:
                item['unitsFulfilled'] += units
        return httpx.Response(
            200,
            json={
                'data': {
                    'contract': self.data['data'][-1],
                    'cargo': self.ship_data['data']['cargo'],
                }
            }
        )

    def reset(self):
        self.data = copy.deepcopy(self.orig_data)
        self.ship_data = copy.deepcopy(self.orig_ship_data)
